from django.contrib.messages import get_messages
from rest_framework import viewsets
from rest_framework.exceptions import PermissionDenied

from drf_messages.conf import MESSAGES_ALLOW_DELETE_UNREAD
from drf_messages.serializers import MessageSerializer


def get_filter_class():
    try:
        from drf_messages.filters import MessageFilterSet
        return MessageFilterSet
    except ImportError:
        return


class MessagesViewSet(viewsets.mixins.ListModelMixin,
                      viewsets.mixins.RetrieveModelMixin,
                      viewsets.mixins.DestroyModelMixin,
                      viewsets.GenericViewSet):
    """
    List, Retrieve and Delete messages for this session.
    """
    serializer_class = MessageSerializer
    search_fields = ("message",)
    ordering_fields = ("level", "read_at", "created")
    filterset_class = get_filter_class()

    def get_queryset(self):
        return get_messages(self.request).get_queryset()

    def check_object_permissions(self, request, obj):
        super(MessagesViewSet, self).check_object_permissions(request, obj)
        # restrict deletion of unread messages.
        if not MESSAGES_ALLOW_DELETE_UNREAD and self.action == "destroy" and obj.read_at is None:
            raise PermissionDenied("Unread messages cannot be deleted.")

    def list(self, request, *args, **kwargs):
        response = super(MessagesViewSet, self).list(request, *args, **kwargs)
        # update last read
        self.get_queryset().mark_read()
        return response



from django.contrib import admin
from django import forms
from django.contrib.messages.storage.base import LEVEL_TAGS

from drf_messages.models import Message, MessageTag


class MessageAdminForm(forms.ModelForm):
    level = forms.ChoiceField(choices=LEVEL_TAGS.items())

    class Meta:
        model = Message
        fields = "__all__"


class MessageTagInline(admin.StackedInline):
    model = MessageTag
    fields = ("text",)
    extra = 0


@admin.register(Message)
class MessageAdmin(admin.ModelAdmin):
    form = MessageAdminForm
    list_display = ("session", "message", "level_tag", "read_at")
    list_filter = ("created", "extra_tags", "level", "read_at", "view", "session")
    readonly_fields = ("session", "created")

    inlines = (MessageTagInline,)




from django.conf import settings


# Allow the deletion of unread messages through DRF view
MESSAGES_ALLOW_DELETE_UNREAD: bool = getattr(settings, "MESSAGES_ALLOW_DELETE_UNREAD", False)
# Automatically read all read messages after request
MESSAGES_DELETE_READ: bool = getattr(settings, "MESSAGES_DELETE_READ", False)
# Use request session for storing messages
MESSAGES_USE_SESSIONS: bool = getattr(settings, "MESSAGES_USE_SESSIONS", False)



# Generated by Django 3.1.5 on 2021-02-13 00:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sessions', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('view', models.CharField(blank=True, default='', help_text='The view where the message was submitted from.', max_length=64)),
                ('message', models.CharField(blank=True, help_text='The actual text of the message.', max_length=1024)),
                ('level', models.IntegerField(help_text='An integer describing the type of the message.')),
                ('read_at', models.DateTimeField(blank=True, default=None, help_text='When the message was read.', null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('session', models.ForeignKey(blank=True, default=None, help_text='The session where the message was submitted to.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='sessions.session')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='MessageTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.CharField(help_text='Custom tags for the message.', max_length=128)),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extra_tags', to='drf_messages.message')),
            ],
        ),
    ]



from rest_framework.routers import DefaultRouter

from drf_messages.views import MessagesViewSet

router = DefaultRouter()
router.register("", MessagesViewSet, "messages")


app_name = "drf_messages"
urlpatterns = [
    *router.urls,
]



from functools import lru_cache

from django.contrib.messages.storage.base import Message as DjangoMessage, BaseStorage

from drf_messages import logger
from drf_messages.conf import MESSAGES_DELETE_READ, MESSAGES_USE_SESSIONS
from drf_messages.models import Message


class DBStorage(BaseStorage):
    """
    Message storage backend to persistent messages storage in the database, with relation to the request's session.
    When no session is provided, it fallbacks to a temporary storage in memory.
    """

    def __init__(self, request, *args, **kwargs):
        super(DBStorage, self).__init__(request, *args, **kwargs)
        # fallback to non persistent message storage when no session is available
        if MESSAGES_USE_SESSIONS:
            self._fallback = not bool(hasattr(request, "session") and request.session.session_key)
        else:
            self._fallback = not bool(hasattr(request, "user") and request.user.is_authenticated)

    def get_queryset(self):
        """
        Get queryset of all messages for that request session.
        :return: MessageQuerySet object
        """
        return Message.objects.with_context(self.request)

    def get_unread_queryset(self):
        """
        Get queryset of unread messages for that request session.
        :return: MessageQuerySet object
        """
        return self.get_queryset().filter(read_at__isnull=True)

    def __iter__(self):
        if self._fallback:
            self.used = True
            yield from self._queued_messages
        else:
            # parse to Django original Message objects
            for message in self.get_unread_queryset().prefetch_related("extra_tags"):
                yield DjangoMessage(
                    level=message.level,
                    message=message.message,
                    extra_tags=str(list(message.extra_tags.values_list("text", flat=True)))
                )

            # update last read
            self.get_queryset().mark_read()

    def __contains__(self, item: DjangoMessage):
        if self._fallback:
            return item in self._queued_messages
        else:
            return self.get_unread_queryset().filter(message=item.message, level=item.level).exists()

    @lru_cache
    def __len__(self):
        if self._fallback:
            return len(self._queued_messages)
        else:
            return self.get_unread_queryset().count()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.get_queryset().mark_read()

    def _store(self, messages, response, *args, **kwargs):
        # messages are save on-demand when is created
        return []

    def _get(self, *args, **kwargs):
        if self._fallback:
            # avoid potential infinite loop, as the super method of __iter__ calls _get itself.
            return [], True
        else:
            return list(self.__iter__()), True

    def add(self, level: int, message: str, extra_tags=''):
        if self._fallback:
            # save messaged to temporary storage in memory
            self._queued_messages.append(DjangoMessage(level, message, extra_tags=extra_tags))
        else:
            if message and int(level) >= self.level:
                Message.objects.create_message(self.request, message, level, extra_tags=extra_tags)
            else:
                if not message:
                    logger.debug(f"Skip message creation due to an empty string. ({message})")
                elif level < self.level:
                    logger.debug(f"Skip message creation due to the level being too low ({level} / {self.level}.")

    def update(self, response):
        # delete already read messages
        if MESSAGES_DELETE_READ and self.used and not self._fallback:
            count, _ = self.get_queryset().filter(read_at__isnull=False).delete()
            logger.info(f"Cleared {count} messages for session {self.request.session}")

    def __str__(self):
        messages = [m.message for m in self.__iter__()]
        self.used = False
        return str(messages)

    def __repr__(self):
        return self.__str__()



from django.contrib.auth import get_user_model
from django.contrib.messages import get_messages
from django.contrib.messages.storage.base import LEVEL_TAGS
from django.contrib.sessions.models import Session
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property

from drf_messages import logger
from drf_messages.conf import MESSAGES_USE_SESSIONS


class MessageQuerySet(models.QuerySet):

    def __init__(self, model=None, query=None, using=None, hints=None, request_context=None):
        super(MessageQuerySet, self).__init__(model=model, query=query, using=using, hints=hints)
        self.request_context = request_context

    def _clone(self):
        # pass request context on clone
        c = super(MessageQuerySet, self)._clone()
        c.request_context = self.request_context
        return c

    def mark_read(self):
        """
        Mark any unread messages as read now.
        :return: Number of messages updated
        """
        # mark that messages have been read from the request
        result = self.filter(read_at__isnull=True).update(read_at=timezone.now())
        logger.debug(f"Marked {result} messages as read for session {self.request_context.session.session_key}")
        if result > 0 and self.request_context:
            storage = get_messages(self.request_context)
            if storage is not None:
                storage.used = True
            else:
                logger.warning("Message storage is None. Make sure to include "
                               "\"'django.contrib.messages.middleware.MessageMiddleware'\" in the MIDDLEWARE setting.")
        return result


class MessageManager(models.Manager):

    def with_context(self, request):
        """
        Filter only messages for a request context:
        When MESSAGES_USE_SESSIONS, messages for that session or without a session specified,
        otherwise messages from all sessions.
        """
        queryset = MessageQuerySet(self.model, using=self._db, request_context=request).filter(
            user=request.user if hasattr(request, "user") and request.user.is_authenticated else None
        )
        if MESSAGES_USE_SESSIONS:
            return queryset.filter(Q(session__session_key=request.session.session_key) | Q(session__isnull=True))
        else:
            return queryset

    def _create_extra_tags(self, message, extra_tags):
        """
        Create message tags from list or string.
        :param message: The message to attach the tags to.
        :param extra_tags: String or List of string tags.
        :return: None
        """
        if isinstance(extra_tags, (list, tuple, set)):
            MessageTag.objects.bulk_create((
                MessageTag(message=message, text=tag)
                for tag in extra_tags
            ))
        else:
            MessageTag.objects.create(message=message, text=str(extra_tags))

    def create_message(self, request, message, level, extra_tags=None):
        """
        Create a new message to the database.
        :param request: Request context.
        :param message: Text body of the message.
        :param level: Integer describing the type of the message.
        :param extra_tags: One or more tags to attach to the message.
        :return: Message object.
        """
        # extract session
        if hasattr(request, "session") and request.session.session_key:
            session = Session.objects.get(session_key=request.session.session_key)
        else:
            session = None
        # create message
        message_obj = self.create(
            user=request.user,
            session=session,
            view=request.resolver_match.view_name if request.resolver_match else '',
            message=message,
            level=level,
        )
        # create extra tags
        if extra_tags:
            self._create_extra_tags(message_obj, extra_tags)

        return message_obj

    def create_user_message(self, user, message, level, extra_tags=None):
        """
        Create a new message to the database.
        :param user: User object (from settings.AUTH_USER_MODEL).
        :param message: Text body of the message.
        :param level: Integer describing the type of the message.
        :param extra_tags: One or more tags to attach to the message.
        :return: Message object.
        """
        # create message
        message_obj = self.create(
            user=user,
            message=message,
            level=level,
        )
        # create extra tags
        if extra_tags:
            self._create_extra_tags(message_obj, extra_tags)

        return message_obj


class MessageTag(models.Model):
    message = models.ForeignKey("drf_messages.Message", on_delete=models.CASCADE, related_name="extra_tags")

    text = models.CharField(max_length=128, help_text="Custom tags for the message.")

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text


class Message(models.Model):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE, related_name="messages")
    session = models.ForeignKey(Session, on_delete=models.CASCADE, null=True, blank=True, default=None,
                                related_name="messages", help_text="The session where the message was submitted to.")
    view = models.CharField(max_length=64, blank=True, default="",
                            help_text="The view where the message was submitted from.")

    message = models.CharField(max_length=1024, blank=True, help_text="The actual text of the message.")
    level = models.IntegerField(help_text="An integer describing the type of the message.")

    read_at = models.DateTimeField(blank=True, null=True, default=None, help_text="When the message was read.")

    created = models.DateTimeField(auto_now_add=True)

    objects = MessageManager()

    class Meta:
        ordering = ["-created"]

    @cached_property
    def level_tag(self):
        return LEVEL_TAGS.get(self.level, '')

    def __str__(self):
        return self.message

    def __repr__(self):
        return self.message



from django.contrib.messages.storage.base import LEVEL_TAGS
from django_filters import FilterSet, BooleanFilter, TypedChoiceFilter, CharFilter, DateTimeFromToRangeFilter

from drf_messages.models import Message

REVERSED_LEVEL_TAGS = {
    v: k for k, v in LEVEL_TAGS.items()
}


class MessageFilterSet(FilterSet):
    unread = BooleanFilter(field_name="read_at", lookup_expr="isnull", label="unread")
    extra_tags = CharFilter(field_name="extra_tags__text")
    level_tag = TypedChoiceFilter(choices=zip(LEVEL_TAGS.values(), LEVEL_TAGS.values()), lookup_expr="gte",
                                  field_name="level", label="level_tag", coerce=lambda k: REVERSED_LEVEL_TAGS.get(k))
    read = DateTimeFromToRangeFilter(field_name="read_at", label="read")
    created = DateTimeFromToRangeFilter()

    class Meta:
        model = Message
        fields = ("unread", "level_tag", "level", "extra_tags", "view", "read", "created")



from django.contrib.messages.storage.base import LEVEL_TAGS
from rest_framework import serializers

from drf_messages.models import Message, MessageTag


class MessageSerializer(serializers.ModelSerializer):
    extra_tags = serializers.SlugRelatedField(slug_field="text", queryset=MessageTag.objects.all(), many=True)
    level = serializers.ChoiceField(choices=tuple(LEVEL_TAGS.items()))
    level_tag = serializers.ChoiceField(choices=tuple(LEVEL_TAGS.values()))

    class Meta:
        model = Message
        fields = ("id", "message", "level", "level_tag", "extra_tags", "view", "read_at", "created")
